package apt;

import java.awt.*;
import java.applet.*;

/**
 *
 * The Display class holds the global display settings for
 * the various apartment components. A Display object can
 * be initialized using the init() method, which takes an Applet
 * and reads the settings from the applet's parameters.
 * There is also a global "defaultDisplay" object that is a convenient
 * way for objects to inherit a default set of settings.
 *
 */
public class Display
{

    // Fonts
    public Font smallFont=new Font("Helvetica", Font.PLAIN, 10);
    public Font labelFont=new Font("Helvetica", Font.PLAIN, 11);
    public Font typeFont=new Font("Courier", Font.BOLD, 14);
    public Font titleFont=labelFont;//new Font("Helvetica", Font.BOLD, 14);
    public Font hugeFont=new Font("Helvetica", Font.BOLD, 24);
    public Font roomFont=hugeFont;
    public Font wordFont=labelFont;
    public Font plainFont=labelFont;
    
    // FontMetrics
    private static FontMetrics metrics(Font font)
        {return Toolkit.getDefaultToolkit().getFontMetrics(font);}
    
    public FontMetrics labelFontMetrics=metrics(labelFont);
    public FontMetrics plainFontMetrics=metrics(plainFont);
    public FontMetrics typeFontMetrics=metrics(typeFont);
    public FontMetrics roomFontMetrics=metrics(roomFont);
    public FontMetrics titleFontMetrics=metrics(titleFont);
    public FontMetrics hugeFontMetrics=metrics(hugeFont);
    public FontMetrics smallFontMetrics=metrics(smallFont);
    
    // Colors
    public Color roomNameColor=Color.yellow;
    public Color[] palette=new Color[256];
    public Color ambientColor;
    public Color recognizedWordColor;
    public Color planColor;
    public Color emptyColor=Color.white;
    public Color labelColor=Color.black;
    public Color lineColor=new Color(0x003399);
    public Color markColor=new Color(0xff9933);
    public Color visitedColor=new Color(235,214,137);
    public Color snapColor=new Color(0xff9933);
    public int palestLevel=225;
    
    // Layout
    public int cursorY=15; // measured from bottom.
    public int rightMargin=114;
    public int buttonHeight=21;
    public int buttonPad=7;
    public int instructionY=30;
    public int titleY=19;
    public int instructionX=20;
    public int createButtonY=cursorY; // measured from bottom.
    public int inputButtonY=130; // measured from bottom.
    public int cityButtonY=50; // measured from bottom.
    public int axisButtonY=instructionY//-labelFontMetrics.getAscent()
                           +(buttonHeight+buttonPad)-5;
    
    // Names of rooms.
    public static final String[] roomNames=
    {
        "window",
        "hall",
        "bathroom",
        "office",
        "living",
        "library",
        "media",
        "dining",
        "foyer",
        "kitchen",
        "bedroom",
        "closet"
    };
    
    
    // pointiness, roughness, spacing.
    // default = 3,2,1
    public static final RoomDescription[] roomDescription=
    {
        new RoomDescription(2,2,.7), // "window",
        RoomDescription.normal, // "hall",
        RoomDescription.normal, // "bathroom",
        new RoomDescription(16,2,RoomDescription.normal.spacing), // "office",
        RoomDescription.normal, // "living",
        RoomDescription.normal, // "library",
        RoomDescription.normal, // "media",
        RoomDescription.normal, // "dining",
        new RoomDescription(2,2,.7), // "foyer",
        RoomDescription.normal, // "kitchen",
        RoomDescription.normal, // "bedroom",
        RoomDescription.normal  // "closet"
    };
    
    public static String modifier(double a)
    {
        if (a<.2) return "very low";
        if (a<.4) return "low";
        if (a<.6) return "medium";
        if (a<.8) return "high";
        return "very high";
    }
    
    
    // Used to refer to window in Plan3D and ApartmentConstructor, among others.
    public static final int WINDOW=0;
    
    // Names of axes corresponding to rooms.
    public String[] axisNames={"Open", "Move", "Body",
        "Work","Social","Truth","Art",
        "Glamour","Change", "Food", "Intimacy", "Secret"};
        
    // Pairings of axes to use in city plot.
    public int[][] pair=
    {
        {10,5}, {0,4}, {11,2}, {3,6}, {7,9}, {8,1}
    };
    
    public String[][][] neighborhood=
    {
        // intimacy/truth
        {
            {"Madison Ave.", null, "Red-light district"}, 
            {null, null, null}, 
            {"Courthouse", "College Street", "Home"}
        },
        
        // open/social
        {
            {"Wall Street", null, "Grand Central Sta."}, 
            {null, null, null}, 
            {"Upper East Side", null, "Times Square"}
        },
        
        // body/secret
        {
            {"Upper West Side", null, null}, 
            {null, null, null}, 
            {"Hospital", null, "Graveyard"}
        },
        
        // art/practical
        {
            {"Golf Course", null, null}, 
            {null, null, null}, 
            { "Greenwich Village", null, "Brooklyn Bridge"}
        },
        
        // glamour/nourish
        {
            {"Hard Rock Cafe", null, "Garment District."}, 
            {null, null, null}, 
            {"Diner", null, null}
        },
        
        // moving/changing
        {
            {"City Hall", null, "Stock Exchange"}, 
            {null, null, null}, 
            {"Subway Station", null, "Riverfront"}
        },
        
    };
    
    // Default display settings.
    private static Display defaultDisplay=new Display();
    

    /**
     *
     * @return The correct title color.
     *
     */
    public Color getTitleColor()
    {
        return palette[170];
    }


    /**
     *
     * @return The correct background color.
     *
     */
    public Color bgcolor()
    {
        return palette[255];
    }

    /**
     *
     * @return The correct midrange color.
     *
     */
    public Color midcolor()
    {
        return palette[128];
    }

    /**
     *
     * @return The correct foreground color.
     *
     */
    public Color fgcolor()
    {
        return palette[0];
    }
    
    /**
     *
     * Which instructions to show on screen for apartment constructor,
     * given the number of sentences the user has typed into
     * the constructor.
     *
     * @param Number of sentences typed into constructor screen.
     * @return Any instructions to display on screen.
     */
    public String constructionHelp(int stage)
    {
        switch (stage)
        {
            case 0: return "To begin construction, type a sentence and hit 'return'.";
            case 1: return "As you type more sentences, your apartment will take shape...";
            default: return "";
        }
    }
    
    /**
     *
     * Return the default display object.
     * Note that this object must be initialized by someone!
     * If not, you'll probably see a great many NullPointerExceptions.
     *
     * @return a Display object containing default settings.
     *
     */
    public static Display getDefaultDisplay()
    {
        return defaultDisplay;
    }
    
    /**
     *
     * Initialize by reading the params of the given applet.
     *
     * @param applet the Applet whose params should be read.
     *
     */
    public void init(Applet applet)
    {
        Color bg=paramColor(applet, "bgcolor");
        Color mid=paramColor(applet, "midcolor");
        Color fg=paramColor(applet, "fgcolor");

        int i;
        for (i=0; i<128; i++)
            palette[i]=interpolate(fg, mid, i, 128);
        for (i=0; i<128; i++)
            palette[128+i]=interpolate(mid, bg, i, 127);

        planColor=palette[244];
        ambientColor=palette[220];
        recognizedWordColor=palette[0];
        
    }
    

    private Color paramColor(Applet applet, String p)
    {
        return new Color(Integer.parseInt(applet.getParameter(p),16));
    }

    private Color interpolate(Color c1, Color c2, int i, int n)
    {
        int r=interpolate(c1.getRed(), c2.getRed(), i,n);
        int g=interpolate(c1.getGreen(), c2.getGreen(), i,n);
        int b=interpolate(c1.getBlue(), c2.getBlue(), i,n);
        return new Color(r,g,b);
    }

    private int interpolate(int a, int b, int i, int n)
    {
        return (a*(n-i)+b*i)/n;
    }
    
}