package apt;

import apt.plan2d.*;
import apt.plan3d.*;

import java.awt.*;
import java.applet.*;
import java.net.*;
import java.util.*;
import java.io.*;

/**
 *
 * The InputScreen3D handles a 2D display and
 * a 3D display simultaneously.
 *
 */
public class InputScreen3D extends Applet implements Runnable
{
    /** ID for an event saying that an apartment was saved. */
    public static int APARTMENT_SAVED=9999;

    /** ID for an event saying that user wants to exit this view. */
    public static int EXIT=8888;

    /** ID for an event saying that user wants 3d. */
    public static int VIEW3D=7777;

    // Interactive construction panel.
    private ApartmentConstructor constructor;

    // 3D display.
    private GridDisplay display3D;

    // 3D controller.
    private DisplayCommandBroadcaster broadcaster=new DisplayCommandBroadcaster();


    /** 3D builder. */
    private Plan3D builder=new Plan3D();

    /** Server interface. */
    private ServerGateway server;

    // "Save" button.
    private OutlineButton saveButton=new OutlineButton("Save");

    // "Clear" button.
    private OutlineButton clearButton=new OutlineButton("Demolish");

    // "City" button.
    private OutlineButton cityButton=new OutlineButton("Back to City");



    /**
     *
     * This version of init() is used only when this is
     * called as a free-standing applet.
     *
     */
    public void init()
    {
        boolean editable="yes".equals(getParameter("editable"));
        Display.getDefaultDisplay().init(this);
        init(size().width, size().height, editable);
    }

    /**
     *
     * This version of the init() method takes
     * a specific size in parameters, so that
     * initialization can take place before
     * the applet is sized for the screen.
     *
     * @param w the desired width.
     * @param h the desired height.
     * @param editable whether the floor plan is editable.
     *
     */
    public void init(int w, int h, boolean editable)
    {
        boolean has3D=true;
        int constructorHeight=has3D ? h/2 : h;
        Display display=Display.getDefaultDisplay();
        int right=display.rightMargin;
        setLayout(null);
        setBackground(Color.white);

        int bh=display.buttonHeight, bp=bh+display.buttonPad, bx=w-right;
        add(saveButton, bx,h-2*bp, right,bh);
        add(clearButton, bx,h-bp, right, bh);
        add(cityButton, bx,h-4*bp, right,bh);


        try
        {

            // read words that define the rooms.
            RoomDefinition roomDefinition=new RoomDefinition();
            URL u=new URL(getDocumentBase(), getParameter("room url"));
            roomDefinition.readDefinition(u.openStream());

            // Read seed for random number generator.
            long seed;
            String seedString=getParameter("seed");
            if (seedString!=null)
                seed=Long.parseLong(seedString);
            else
                seed=System.currentTimeMillis();

            // Add apartment constructor.
            constructor=new ApartmentConstructor();
            add(constructor);
            constructor.reshape(0,0,w-right, constructorHeight);

            // Read any initial words via params.
            String words=getParameter("initial words");
            if (words!=null)
            {
                StringTokenizer t=new StringTokenizer(words);
                while(t.hasMoreElements())
                {
                    String wd=t.nextToken();
                    constructor.addInitialWord(wd);
                }
            }

            // Initialize the apartment constructor.
            constructor.init(seed, roomDefinition, w-right, constructorHeight);


            // UNCOMMENT THIS BLOCK TO PUT 3D DISPLAY IN.
            // YOU WILL ALSO HAVE TO CHANGE THE SIZE OF THE
            // APPLET AND THE "CONSTRUCTOR" OBJECT.

            // Add 3D display.
            display3D=new GridDisplay();
            add(display3D);
            display3D.reshape(0,h/2,w-right,h/2);
            display3D.setColor(new Color(0x9999ff));
            display3D.init();

            // Register 3D object as listener.
            addController(display3D);


        } catch (Exception e)
        {
            e.printStackTrace(System.err);
        }

    }

    public void setServer(ServerGateway server)
    {
        this.server=server;
    }

    /**
     *
     * Set whether this apartment is saveable or not.
     *
     * @param saveable whether this apartment is saveable or not.
     *
     */
    public void setSaveable(boolean saveable)
    {
        if (saveable)
        {
            saveButton.enable();
            //cityButton.setLabel("Discard and go to city");
        }
        else
        {
            saveButton.disable();
            //cityButton.setLabel("Back to city");
        }
    }

    /**
     *
     * Clear the contents of the constructor.
     *
     */
    public void clear()
    {
        constructor.clear();
        constructor.requestFocus();
        setSaveable(true);
    }

    /**
     *
     * We override the show() method so that the main
     * constructor panel can request the focus.
     * Otherwise, the user sometimes types and is
     * baffled by the fact that nothing happens.
     *
     */
    public void show()
    {
        constructor.erase();
        super.show();
        constructor.requestFocus();
    }

    /**
     *
     * This is just a helper method for layout.
     *
     */
    private void add(Component c, int x, int y, int w, int h)
    {
        add(c);
        c.setFont(Display.getDefaultDisplay().plainFont);
        c.reshape(x,y,w,h);
    }

    /**
     *
     * Override to print warning message for null param.
     *
     */
    public String getParameter(String s)
    {
        String p=super.getParameter(s);
        if (p==null)
            System.out.println("Warning: null param: "+s);
        return p;
    }

    /**
     *
     * Handle events signalling that the apartment has been
     * updated or cleared.
     *
     */
    public boolean handleEvent(Event e)
    {
        if (e.arg==ApartmentEvent.UPDATE)
        {
            Hashtable info3D=builder.translate2Dto3D(
                constructor.getRooms(), constructor.getClassifiedWords());
            broadcaster.update(info3D);
            /*
            try
            {
                apt.vrml.WordSurfaceIO.writeAll(info3D, System.out);
            }
            catch (Exception ex)
            {
                System.out.println(ex);
            }
            */
        }
        else if (e.arg==ApartmentEvent.CLEAR)
            broadcaster.clearSurfaces();

        return super.handleEvent(e);
    }

    /**
     *
     * Handle events from buttons for clear, save, and exit to city.
     *
     */
    public boolean action(Event e, Object arg)
    {
        if (e.target==saveButton)
        {
            saveButton.disable();
            new Thread(this).start();
        }
        else if (e.target==clearButton)
            clear();
        else if (e.target==cityButton)
            postEvent(new Event(this, EXIT, null));

        return true;
    }

    /**
     *
     * Set the apartment for this screen. If the specified apartment is
     * null, then the constructor will be initialized with
     * an empty apartment.
     *
     * @param apartment the apartment object to use.
     *
     */
    public void setApartment(Apartment apartment)
    {
        constructor.setApartment(apartment);
    }

    /**
     *
     * Get summary of current apartment that's been
     * entered by the user.
     *
     * @return an Apartment with all data
     *         needed to describe this apartment.
     *
     */
    public Apartment getApartment()
    {
        return constructor.getApartment();
    }

    /**
     *
     * The run() method saves the apartment to server.
     *
     */
    public void run()
    {
        //constructor.setSaving(true);
        constructor.setMessage("Saving apartment, with title: "+
            constructor.getApartment().getTitle());
        try
        {
            Apartment apartment=constructor.getApartment();
            if (apartment==null || apartment.isEmpty())
                System.out.println("Empty apartment. Not saving.");
            else
            {
                server.saveApartment(apartment);
                System.out.println("Saved ok.");
            }
        }
        catch (Exception e)
        {
            System.out.println("Couldn't save.");
            e.printStackTrace();
        }
        postEvent(new Event(this, APARTMENT_SAVED, null));
    }

    /**
     *
     * Add a 3D controller to the list of controllers
     * to receive commands.
     *
     * @param display3D the SurfaceDisplayController to be notified.
     *
     */
    private void addController(SurfaceDisplayController display3D)
    {
        broadcaster.addController(display3D);
    }

    /**
     *
     * Start animation and drawing threads for 2D and 3D displays.
     *
     */
    public synchronized void start()
    {
        stop();
        constructor.start();
        if (display3D!=null)
            display3D.start();
    }

    /**
     *
     * Stop animation and drawing threads for 2D and 3D displays.
     *
     */
    public synchronized void stop()
    {
        if (display3D!=null)
            display3D.stop();
        constructor.stop();
    }
}
