package apt.plan3d;

import java.util.*;
import space.*;
import apt.plan2d.*;
import apt.*;
import apt.vrml.*;

/**
 *
 * This class converts from a 2D floorplan to a 3D set
 * of grids which cover walls of a 3D structure.
 *
 */
public class Plan3D
{
    // We keep a SpherePoints object for each room.
    // This way the arrangement of items in the room
    // is consistent for each run, rather than changing randomly.
    private SpherePoints[] sphere;
    private long seed=System.currentTimeMillis();
    private int GRID_RADIUS=2;

    // new, improved translation engine.
    public synchronized RoomInfo[] makeRoomInfos(Room[] nodes,
                                           Vector typedWords)
    {
        int n=nodes.length;
        int i;
        
        // Create objects that disperse points.
        if (sphere==null || sphere.length!=nodes.length)
        {
            sphere=new SpherePoints[n];
            Random random=new Random(seed);
            for (i=0; i<n; i++)
                sphere[i]=new SpherePoints(random.nextLong());
        }
        
        // Create ellipsoids corresponding to rooms.
        Ellipsoid[] ellipsoid=new Ellipsoid[n];
        for (i=0; i<n; i++)
        {
            Room r=nodes[i];
            double s=r.getSize();
            if (s<1)
            {
                ellipsoid[i]=new Ellipsoid(-10000,0,0,1,1,1);
                continue;
            }
            double scale=1.2;

            double depth=Math.min(r.w,Math.min(r.h,100+20*s));
            ellipsoid[i]=Ellipsoid.fromBox(r.x-250, -100, r.y-250,
                      r.w*scale, depth*scale, r.h*scale);

        }

        // put points on ellipsoids.
        int fn=typedWords.size();
        Hashtable table=new Hashtable();
        RoomInfo[] rooms=new RoomInfo[n];
        for (i=0; i<n; i++)
        {
            Room r=nodes[i];
            Vector words=new Vector();
            for (int j=0; j<fn; j++)
            {
                TypedWord f=(TypedWord)typedWords.elementAt(j);
                if (f.getType()==i)
                    words.addElement(f);
            }
            int s=words.size();
            double scale=1.2;
            if (s>=1)
            {
                addRoom3D(sphere[i], ellipsoid, i, table, words, rooms, Display.roomNames[i]);
            }
        }
       
        return rooms;        
    }

    /**
    Takes an array of Rooms and a Vector of
    TypedWords, and returns a Hashtable where the keys are
    TypedWords and the values are grids of points that
    represent where the image should go.
    */
    public synchronized Hashtable translate2Dto3D(Room[] nodes,
                        Vector typedWords)
    {
        int n=nodes.length;
        int i;
        
        // Create objects that disperse points.
        if (sphere==null || sphere.length!=nodes.length)
        {
            sphere=new SpherePoints[n];
            Random random=new Random(seed);
            for (i=0; i<n; i++)
                sphere[i]=new SpherePoints(random.nextLong());
        }
        
        // Create ellipsoids corresponding to rooms.
        Ellipsoid[] ellipsoid=new Ellipsoid[n];
        for (i=0; i<n; i++)
        {
            Room r=nodes[i];
            double s=r.getSize();
            if (s<1)
            {
                ellipsoid[i]=new Ellipsoid(-10000,0,0,1,1,1);
                continue;
            }
            double scale=1.2;

            double depth=Math.min(r.w,Math.min(r.h,100+20*s));
            ellipsoid[i]=Ellipsoid.fromBox(r.x-250, -100, r.y-250,
                      r.w*scale, depth*scale, r.h*scale);

        }

        // put points on ellipsoids.
        Hashtable table=new Hashtable();
        int fn=typedWords.size();
        for (i=0; i<n; i++)
        {
            Room r=nodes[i];
            Vector words=new Vector();
            for (int j=0; j<fn; j++)
            {
                TypedWord f=(TypedWord)typedWords.elementAt(j);
                if (f.getType()==i)
                    words.addElement(f);
            }
            int s=words.size();
            double scale=1.2;
            if (s>=1)
            {
                addRoom3D(sphere[i], ellipsoid, i, table, words, null, null);
            }
        }
        return table;
    }


    void addRoom3D(SpherePoints disperser, Ellipsoid[] ellipsoid, int index,
                          Hashtable table, Vector words, RoomInfo[] rooms, String name)
    {
        int n=words.size();
        if (n<=0) return;
        Ellipsoid e=ellipsoid[index];
        RoomInfo room=null;
        if (rooms!=null)
        {
            rooms[index]=new RoomInfo(name);
            room=rooms[index];
            room.setCorners(e.corner(), e.corner().plus(e.dimensions()));
        }


        Point3D[] p=disperser.makeArrangement(n, ellipsoid, index);
        Point3D center=e.center();
        RoomDescription desc=Display.roomDescription[index];
        int i;

        for (i=0; i<n; i++)
        {
            double d=-3;

            for (int j=0; j<n; j++)
            {
                if (i==j) continue;
                double d2=e.surfaceCos(p[i],p[j]);
                if (d2<=d) continue;
                d=d2;
            }

            Point3D unit=p[i].minus(center);
            unit.x/=e.a;
            unit.y/=e.b;
            unit.z/=e.c;
            double angle=.7*(n<3 ? .5*Math.PI : Math.acos(d));
            angle=Math.min(angle, .49*Math.PI);
            Point3D[][] q=patch(unit,GRID_RADIUS,desc.spacing*angle,e,desc.pointiness);
            table.put(words.elementAt(i), q);
            if (room!=null)
            {
                room.addWordSurface(new DefaultWordSurface(
                    ((TypedWord)words.elementAt(i)).getWord(), q));
            }
        }
    }



    private Point3D[][] patch(Point3D p, int k, double angle, Ellipsoid ellipsoid,
                              double exponent)
    {
        /*
        // Debug by creating a rectangle of constant size where patch is.
        Point3D[][] q=new Point3D[2][2];
        double s=5;
        q[0][0]=new Point3D(p.x,p.y,p.z);
        q[1][0]=new Point3D(p.x+s,p.y,p.z);
        q[0][1]=new Point3D(p.x,p.y,p.z+s);
        q[1][1]=new Point3D(p.x+s,p.y,p.z+s);
        */
        
        Point3D a=new Point3D(p.z, 0, -p.x).unit();
        Point3D b=p.cross(a).unit();
        int d=2*k;
        Point3D[][] q=new Point3D[d][d];
        int i,j;

        for (i=-k; i<k; i++)
        {
            double a_factor=-Math.tan(angle*(i+.5)/(k-.5));
            for (j=-k; j<k; j++)
            {
                double b_factor=Math.tan(angle*(j+.5)/(k-.5));
                Point3D offset=a.times(a_factor).plus(b.times(b_factor));
                q[j+k][i+k]=p.plus(offset).unit();
                normalize(q[j+k][i+k],ellipsoid,exponent);
            }
        }
        //exaggerateConvexity(q, angle);
        return q;
    }
    
    /*
    sigh.
    // assumes a square array with even # items on a side.
    private void exaggerateConvexity(Point3D[][] q, double angle)
    {
        int d=q.length;
        int k=d/2;
        
        // compute convexity as (centroid of corners)-(centroid of center).
        Point3D convexity=avg(q[k][k], q[k-1][k], q[k-1][k-1], q[k][k-1])-
                          avg(q[0][0], q[d-1][d-1], q[d-1][0], q[0][d-1]);
                          
        int i,j;

        for (i=-k; i<k; i++)
        {
            for (j=-k; j<k; j++)
            {
                double distance
                q[i][j].translate(convexity.times(factor));
            }
        }
        
    }
    */
    

    private Point3D avg(Point3D a, Point3D b, Point3D c, Point3D d)
    {
        Point3D p= new Point3D(a.x+b.x+c.x+d.x, 
                           a.y+b.y+c.y+d.y,
                           a.z+b.z+c.z+d.z);
        p.scale(.25);
        return p;
    }
   

    private void normalize(Point3D m, Ellipsoid ellipsoid, double q)
    {
        
        double rq=Math.pow(Math.abs(m.x),q)+
                  Math.pow(Math.abs(m.y),q)+
                  Math.pow(Math.abs(m.z),q);
        double r=1/Math.pow(rq, 1/q);
        
        // for debug: use below for automatic spheres.
        //double r=1;
        m.x=ellipsoid.x+ellipsoid.a*m.x*r;
        m.y=ellipsoid.y+ellipsoid.b*m.y*r;
        m.z=ellipsoid.z+ellipsoid.c*m.z*r;
    }



}
