package apt.plan3d;

import java.util.*;
import space.*;

/**
 * 
 * The purpose of this class is to
 * create an arrangement of n points on the unit sphere
 * which are well-dispersed--that is, neighbors are as far
 * apartment as my simple optimization method can make them.
 * <p>
 * It uses a random number generator so that the results vary;
 * hence there is a constructor which takes a seed value, so
 * that if you want consistency you can have it.
 *
 */
public class SpherePoints
{
    private ExtendedRandom random;
    private long seed;

    /**
     *
     * Create a SpherePoints object with a seed
     * corresponding to the current time.
     *
     */
    public SpherePoints()
    {
        this(System.currentTimeMillis());
    }

    /**
     *
     * Create a SpherePoints object with 
     * the specified seed.
     *
     * @param seed the seed for the random number generator.
     *
     */
    public SpherePoints(long seed)
    {
        this.seed=seed;
    }
    
    
    
    
    

    /**
     *
     * Return an arrangement of n points on the unit sphere
     * which are well-dispersed--that is, neighbors are as far
     * apartment as my simple optimization method can make them.
     * <P>
     * This is not a threadsafe method.
     *
     * @param n the number of points to disperse.
     * @return an array of n Point3D objects.
     *
     */
    public Point3D[] makeArrangement(int n, Ellipsoid[] ellipsoid, int index)
    {
        if (n<=0) return new Point3D[0];

        random=new ExtendedRandom(seed);

        Point3D[] p=new Point3D[n];
        int i;

        for (i=0; i<n; i++)
        {
            int safecount=0;
            do
                p[i]=ellipsoid[index].random(random);
            while (++safecount<50 && collides(p[i], ellipsoid, index));
            if (safecount==50)
                System.out.println("punted while making initial points.");
        }

        for (i=0; i<300; i++)
            disperseOnEllipsoid(p, .02, ellipsoid, index);

        return p;

    }
    
    /**
    Does this point collide with any of the given Spheres,
    except for the one with given index?
    */
    private boolean collides(Point3D p, Ellipsoid[] ellipsoid, int index)
    {
        for (int i=ellipsoid.length-1; i>=0; i--)
        {
            if (i==index) continue;
            if (ellipsoid[i].contains(p, 1.2)) 
            {
                //System.out.println("collide!");
                return true;
            }
        }
        return false;
    }

    /**
     *
     * This is the key algorithm for the routine and takes
     * a group of points on the surface of the unit sphere and
     * pushes them away from each other.
     *
     */
    private void disperseOnEllipsoid(Point3D[] p, double delta, Ellipsoid[] ellipsoid, int index)
    {
        int n=p.length;
        Point3D dp=new Point3D(0,0,0);
        for (int i=0; i<n; i++)
        {
            // find the closest point to p[i].
            double d=-3;

            for (int j=0; j<n; j++)
            {
                if (i==j) continue;
                double d2=ellipsoid[index].surfaceCos(p[i],p[j]);//p[i].dot(p[j]);
                if (d2<=d) continue;
                d=d2;
                dp=p[j];
            }
            
            // save old values
            double oldX=p[i].x;
            double oldY=p[i].y;
            double oldZ=p[i].z;

            // push this point away from it.
            p[i].x+=delta*(p[i].x-dp.x);
            p[i].y+=delta*(p[i].y-dp.y);
            p[i].z+=delta*(p[i].z-dp.z);
            ellipsoid[index].projectToSurface(p[i]);
            
            // if collision, forget about it.
            if (collides(p[i], ellipsoid, index))
            {
                p[i].x=oldX;
                p[i].y=oldY;
                p[i].z=oldZ;
            }
        }
    }
    
    
    
    
    
    

    /**
     *
     * Return an arrangement of n points on the unit sphere
     * which are well-dispersed--that is, neighbors are as far
     * apartment as my simple optimization method can make them.
     * <P>
     * This is not a threadsafe method.
     *
     * @param n the number of points to disperse.
     * @return an array of n Point3D objects.
     *
     */
    public Point3D[] makeArrangement(int n)
    {
        if (n<=0) return new Point3D[0];

        random=new ExtendedRandom(seed);

        Point3D[] p=new Point3D[n];
        int i;

        for (i=0; i<n; i++)
            p[i]=random.randomUnit();

        for (i=0; i<300; i++)
            disperseOnUnitSphere(p, .02);

        return p;

    }

    /**
     *
     * This is the key algorithm for the routine and takes
     * a group of points on the surface of the unit sphere and
     * pushes them away from each other.
     *
     */
    private void disperseOnUnitSphere(Point3D[] p, double delta)
    {
        int n=p.length;
        Point3D dp=new Point3D(0,0,0);
        for (int i=0; i<n; i++)
        {
            // find the closest point to p[i].
            double d=-3;

            for (int j=0; j<n; j++)
            {
                if (i==j) continue;
                double d2=p[i].dot(p[j]);
                if (d2<=d) continue;
                d=d2;
                dp=p[j];
            }

            // push this point away from it.
            p[i].x+=delta*(p[i].x-dp.x);
            p[i].y+=delta*(p[i].y-dp.y);
            p[i].z+=delta*(p[i].z-dp.z);
            p[i].scale(1/p[i].norm());
        }
    }

}