package com.bewitched.servlet;

import java.util.*;
import java.io.*;

/**
 * This class is a useful way to store the text
 * or HTML used by a servlet.
 *
 * You construct a TextResource with a file that looks like:
 * <code>
 *         #define favorite search
 *         http://www.google.com
 *
 *         #rem This line is a remark and will be ignored.
 *         #define test
 *         <b>This is a test.</b><p>
 *         I sure hope it work! That would be neat.
 * </code>
 * and then you can access the Strings in the file
 * by saying
 * <code>
 *         myString=myTextResource.get("favorite search");
 * </code>
 *
 * The key of the string goes after "#define".
 * All lines following until the next #define are concatenated to make
 * the text for that key. Leading and trailing whitespace
 * is ignored. Any line beginning with "#rem" is also ignored.
 * All lines before the first #define are ignored.
 * <p>
 * Case is significant for the keywords so, e.g., #REM and #DEFINE
 * are not treated as keywords.
 * <p>
 * The TextResource keeps all the text in RAM, so accessing
 * it is fast, but obviously you would not want to store
 * lots and lots of text this. way.
 */
public class TextResource
{
    private Hashtable strings=new Hashtable();
    private String fileName;
    private final String crlf=""+((char)13)+((char)10);


    /**
    Create a TextResource that takes its text from the given file.
    */
    public TextResource(File file)
    {
        try
        {
            load(new FileReader(file));
            fileName=file.toString();
        }
        catch (Exception e)
        {
            fileName=e.toString();
        }
    }

    /**
    Create a TextResource that takes its text from the given Reader.
    */
    public TextResource(Reader reader)
    {
        load(reader);
    }

    /**
    Create a TextResource from the resource associated with a given class.
    */
    public TextResource(Class c, String resourceName)
    {
        load(c.getResourceAsStream(resourceName));
    }

    /**
    Create a TextResource that takes its text from the given InputStream.
    */
    public TextResource(InputStream in)
    {
        load(in);
    }

    private void load(InputStream in)
    {
        try {load(new InputStreamReader(in));}
        catch (Exception e) {fileName=e.toString();}
    }

    private void load(Reader reader)
    {
        try
        {
            BufferedReader in=new BufferedReader(reader);
            String line=null;
            String current=null;
            StringBuffer b=new StringBuffer();
            while (null != (line=in.readLine()))
            {
                if (line.indexOf("#rem")==0)
                    continue;
                if (line.indexOf("#define")==0)
                {
                    if (current!=null)
                        strings.put(current, b.toString().trim());
                    if (line.length()>7)
                    {
                        current=line.substring(7).trim();
                        b=new StringBuffer();
                    }
                }
                else if (current!=null)
                {
                    b.append(line);
                    b.append(crlf);
                }
            }
            if (current!=null)
            {
                strings.put(current, b.toString().trim());
            }

            in.close();
        }
        catch (Exception e) {fileName=e.toString();}
    }



    /**
    Get the text that corresponds to the given key.

    @param String key
    @return String text for that key.
    */
    public String get(String key)
    {
        if (strings==null)
            return "[no-file-found:"+fileName+"]";
        String value=(String)strings.get(key);
        if (value==null)
            value="[undefined-or-misspelled:"+key+"]";
        return value;
    }

    /**
     *
     * Print to a PrintWriter in HTML form;
     * useful for dumping properties to a servlet
     * for diagnostic info.
     *
     */
    public void printHTML(PrintWriter out)
    {
        Enumeration k=strings.keys();
        while (k.hasMoreElements())
        {
            Object key=k.nextElement();
            out.println("<b>"+key+"</b><br>");
            out.println("<xmp>");
            out.println(strings.get(key));
            out.println("</xmp><p><p>");
        }
    }
}