package space;

import java.awt.*;
import java.util.*;

/**
Panel to display a wireframe 3D scene.
*/
public class WireframePanel extends Panel implements Runnable
{
    private Image buffer;
    private Graphics offscreen;
    private double mx, my, md, distance=15;
    private int cx,cy;
    private Matrix3D base=Matrix3D.I();
    private Vector points=new Vector();
    private Vector segments=new Vector();
    private Viewpoint view;
    private Point3D light=new Point3D(100,0,0);
    private String label="";
    private Color color=Color.white;
    
    private Thread drawing;
    private long frames;
    
    public void setDistance(double distance)
    {
        this.distance=distance;
        makeViewpoint(size().width, size().height);
    }
    
    public synchronized void clear()
    {
        points=new Vector();
        segments=new Vector();
    }
    
    public void stop()
    {
        drawing=null;
    }
    
    public void start()
    {
        (drawing=new Thread(this)).start();
    }
    
    public void setColor(Color color)
    {
        this.color=color;
    }
    
    public Vector getPoints()
    {
        return points;
    }
    
    public synchronized void setLabel(String label)
    {
        this.label=label;
    }
    
    private synchronized long getFrames()
    {
        return frames;
    }
    
    public synchronized void nextFrame()
    {
        frames++;
    }
    
    public void run()
    {
        long currentFrames=-1;
        while (Thread.currentThread()==drawing)
        {
            long tempFrames=getFrames(); 
            if (tempFrames!=currentFrames)
            {
                draw3D(true);
                currentFrames=tempFrames;
            }
            try
            {
                Thread.sleep(50);
            }
            catch (Exception e) {}
        }
    }
    
    synchronized public void addSegment(Segment3D segment)
    {
        segments.addElement(segment);
        addPoint(segment.begin);
        addPoint(segment.end);
    }
    
    synchronized public void addSegment(Point3D p, Point3D q)
    {
        addSegment(new Segment3D(p,q));
    }
    
    synchronized public void addColorSegment(Point3D p, Point3D q, Color c)
    {
        addSegment(new Segment3D(p,q,c));
    }
    
    synchronized private void addPoint(Point3D p)
    {
        if (!points.contains(p))
            points.addElement(p);
    }

    synchronized public void reshape(int x, int y, int w, int h)
    {
        cx=w/2;
        cy=h/2;
        if (offscreen!=null)
            offscreen.dispose();
        buffer=createImage(w,h);
        offscreen=buffer.getGraphics();
        makeViewpoint(w,h);
        super.reshape(x,y,w,h);
        draw3D(true);
    }
    
    private void makeViewpoint(int w, int h)
    {
        view=new Viewpoint(Matrix3D.I(), light,
                           distance,Math.max(w,h),cx,cy);
    }

    synchronized private void draw3D(boolean drawNow)
    {
        if (offscreen==null) return;
        int w=size().width;
        int h=size().height;
        offscreen.setColor(getBackground());
        offscreen.setColor(Color.white);
        offscreen.fillRect(0,0,w,h);
        offscreen.setColor(color);
        
        int i;
        for (i=points.size()-1; i>=0; i--)
        {
            Point3D p=(Point3D)points.elementAt(i);
            p.projection(view);
        }
        
        int n=segments.size();
        
        for (i=n-1; i>=0; i--)
        {
            Segment3D s=(Segment3D)segments.elementAt(i);
            offscreen.setColor(s.color);
            offscreen.drawLine(s.begin.projection.x,
                               s.begin.projection.y,
                               s.end.projection.x,
                               s.end.projection.y);
        }
        
        offscreen.drawString(label,5,30);
        
        if (drawNow) repaint();
    }

    public void update(Graphics g) {paint(g);}

    synchronized public void paint(Graphics g)
    {
        if (buffer!=null)
        {
            g.drawImage(buffer, 0,0, this);
        }
    }

    public boolean mouseDown(Event e, int x, int y)
    {
        mx=x-cx;
        my=cy-y;
        base=Matrix3D.rotateXY(mx,my,100).inverse().times(view.angle);
        return true;
    }

    public boolean mouseDrag(Event e, int x, int y)
    {
        double a=x-cx;
        double b=cy-y;
        synchronized (this)
        {
            view.rotateObjectsTo(
                Matrix3D.rotateXY(a,b,100).times(base), light);
        }
        nextFrame();
        return true;
    }
}